# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: service.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)




DESCRIPTOR = _descriptor.FileDescriptor(
  name='service.proto',
  package='zmp',
  serialized_pb='\n\rservice.proto\x12\x03zmp\"C\n\x12ServiceRequestImpl\x12\x0f\n\x07service\x18\x01 \x01(\t\x12\x0c\n\x04\x61rgs\x18\x02 \x01(\x0c\x12\x0e\n\x06kwargs\x18\x03 \x01(\x0c\"f\n\x13ServiceResponseImpl\x12\x0f\n\x07service\x18\x01 \x01(\t\x12\x10\n\x08response\x18\x02 \x01(\x0c\x12,\n\texception\x18\x03 \x01(\x0b\x32\x19.zmp.ServiceExceptionImpl\"N\n\x14ServiceExceptionImpl\x12\x10\n\x08\x65xc_type\x18\x01 \x01(\t\x12\x11\n\texc_value\x18\x02 \x01(\x0c\x12\x11\n\ttraceback\x18\x03 \x01(\x0c\x32K\n\x0bServiceImpl\x12<\n\x07Service\x12\x17.zmp.ServiceRequestImpl\x1a\x18.zmp.ServiceResponseImpl')




_SERVICEREQUESTIMPL = _descriptor.Descriptor(
  name='ServiceRequestImpl',
  full_name='zmp.ServiceRequestImpl',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='service', full_name='zmp.ServiceRequestImpl.service', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='args', full_name='zmp.ServiceRequestImpl.args', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='kwargs', full_name='zmp.ServiceRequestImpl.kwargs', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=22,
  serialized_end=89,
)


_SERVICERESPONSEIMPL = _descriptor.Descriptor(
  name='ServiceResponseImpl',
  full_name='zmp.ServiceResponseImpl',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='service', full_name='zmp.ServiceResponseImpl.service', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='response', full_name='zmp.ServiceResponseImpl.response', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='exception', full_name='zmp.ServiceResponseImpl.exception', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=91,
  serialized_end=193,
)


_SERVICEEXCEPTIONIMPL = _descriptor.Descriptor(
  name='ServiceExceptionImpl',
  full_name='zmp.ServiceExceptionImpl',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='exc_type', full_name='zmp.ServiceExceptionImpl.exc_type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='exc_value', full_name='zmp.ServiceExceptionImpl.exc_value', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='traceback', full_name='zmp.ServiceExceptionImpl.traceback', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=195,
  serialized_end=273,
)

_SERVICERESPONSEIMPL.fields_by_name['exception'].message_type = _SERVICEEXCEPTIONIMPL
DESCRIPTOR.message_types_by_name['ServiceRequestImpl'] = _SERVICEREQUESTIMPL
DESCRIPTOR.message_types_by_name['ServiceResponseImpl'] = _SERVICERESPONSEIMPL
DESCRIPTOR.message_types_by_name['ServiceExceptionImpl'] = _SERVICEEXCEPTIONIMPL

class ServiceRequestImpl(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _SERVICEREQUESTIMPL

  # @@protoc_insertion_point(class_scope:zmp.ServiceRequestImpl)

class ServiceResponseImpl(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _SERVICERESPONSEIMPL

  # @@protoc_insertion_point(class_scope:zmp.ServiceResponseImpl)

class ServiceExceptionImpl(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _SERVICEEXCEPTIONIMPL

  # @@protoc_insertion_point(class_scope:zmp.ServiceExceptionImpl)


# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

# To allow python to run these tests as main script
import functools
import multiprocessing
import sys
import os
import threading

import types
from random import randint

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..')))

import time
import pyzmp

import pytest
# http://pytest.org/latest/contents.html
# https://github.com/ionelmc/pytest-benchmark

# TODO : PYPY
# http://pypy.org/

# TODO : Test Node exception : correctly transmitted, node still keeps spinning...


@pytest.mark.timeout(5)
def test_node_termination():
    """Checks that a node can be shutdown without being started and indicate that it never ran"""
    n1 = pyzmp.Node()
    assert not n1.is_alive()
    exitcode = n1.shutdown()  # shutdown should have no effect here (if not started, same as noop )
    assert exitcode is None  # exitcode should be None (process didn't start and didn't stop so no exit code)
    assert not n1.is_alive()


@pytest.mark.timeout(5)
def test_node_creation_termination():
    """Checks that a node can be started and shutdown and indicate that it ran successfully"""
    n1 = pyzmp.Node()
    assert not n1.is_alive()
    n1.start()
    assert n1.is_alive()
    exitcode = n1.shutdown()
    assert exitcode == 0  # default node should spin without issues
    assert not n1.is_alive()


@pytest.mark.timeout(5)
def test_node_timeout_creation_termination():
    """Checks that a node can be started with timeout and shutdown and indicate that it ran successfully"""
    n1 = pyzmp.Node()
    assert not n1.is_alive()
    started = n1.start(1)
    assert started
    assert n1.is_alive()
    exitcode = n1.shutdown()
    assert exitcode == 0
    assert not n1.is_alive()


# @nose.SkipTest  # to help debugging ( FIXME : how to programmatically start only one test - maybe in fixture - ? )
@pytest.mark.timeout(5)
def test_node_double_creation_termination():
    """Checks that a node can be started twice and shutdown and indicate that it ran successfully"""
    n1 = pyzmp.Node()
    assert not n1.is_alive()
    n1.start()
    assert n1.is_alive()
    n1.start()  # this shuts down properly and restart the node
    assert n1.is_alive()

    exitcode = n1.shutdown()
    assert exitcode == 0
    assert not n1.is_alive()


# @nose.SkipTest  # to help debugging ( FIXME : how to programmatically start only one test - maybe in fixture - ? )
@pytest.mark.timeout(5)
def test_node_timeout_double_creation_termination():
    """Checks that a node can be started twice with timeout and shutdown and indicate that it ran successfully"""
    n1 = pyzmp.Node()
    assert not n1.is_alive()
    assert n1.start(1)
    assert n1.is_alive()
    assert n1.start(1)  # this shuts down and restart the node
    assert n1.is_alive()

    exitcode = n1.shutdown()
    assert exitcode == 0
    assert not n1.is_alive()


@pytest.mark.timeout(5)
def test_node_creation_double_termination():
    """Checks that a node can be started and shutdown twice and indicate that it ran successfully"""
    n1 = pyzmp.Node()
    assert not n1.is_alive()
    n1.start()
    assert n1.is_alive()
    exitcode = n1.shutdown()
    assert exitcode == 0
    assert not n1.is_alive()
    exitcode = n1.shutdown()
    assert exitcode == 0  # the exit code is still 0 since we didn't restart...
    assert not n1.is_alive()


@pytest.mark.timeout(5)
def test_node_creation_args():
    """Checks that a node can be passed an argument using inheritance"""
    ns = multiprocessing.Manager().Namespace()
    ns.arg = 42

    class TestArgNode(pyzmp.Node):
        def update(self, *args, **kwargs):
            ns.arg -= args[0]
            return ns.arg

    n1 = TestArgNode(args=(ns.arg,))
    assert not n1.is_alive()
    n1.start()
    assert n1.is_alive()
    exitcode = n1.shutdown()
    assert exitcode == 0
    assert not n1.is_alive()

    assert ns.arg == 0


@pytest.mark.timeout(5)
def test_node_creation_args_delegate():
    """Checks that a node can be passed an argument using delegation"""
    ns = multiprocessing.Manager().Namespace()
    ns.arg = 42

    def arguser(fortytwo, **kwargs):  # kwargs is there to accept extra arguments nicely (timedelta)
        ns.arg -= fortytwo
        return ns.arg

    n1 = pyzmp.Node(args=(ns.arg,), target=arguser)
    assert not n1.is_alive()
    n1.start()
    assert n1.is_alive()
    exitcode = n1.shutdown()
    assert exitcode == 0
    assert not n1.is_alive()

    assert ns.arg == 0


@pytest.mark.timeout(5)
def test_node_creation_kwargs():
    """Checks that a node can be passed a keyword argument using inheritance"""
    ns = multiprocessing.Manager().Namespace()
    ns.kwarg = 42

    class TestKWArgNode(pyzmp.Node):
        def update(self, *args, **kwargs):
            ns.kwarg -= kwargs.get('intval')
            return ns.kwarg

    n1 = TestKWArgNode(kwargs={'intval': ns.kwarg, })
    assert not n1.is_alive()
    n1.start()
    assert n1.is_alive()
    exitcode = n1.shutdown()
    assert exitcode == 0
    assert not n1.is_alive()

    assert ns.kwarg == 0


@pytest.mark.timeout(5)
def test_node_creation_kwargs_delegate():
    """Checks that a node can be passed a keyword argument using delegation"""
    ns = multiprocessing.Manager().Namespace()
    ns.kwarg = 42

    def kwarguser(intval, **kwargs):  # kwargs is there to accept extra arguments nicely (timedelta)
        ns.kwarg -= intval
        return ns.kwarg

    n1 = pyzmp.Node(kwargs={'intval': ns.kwarg, }, target=kwarguser)
    assert not n1.is_alive()
    n1.start()
    assert n1.is_alive()
    exitcode = n1.shutdown()
    assert exitcode == 0
    assert not n1.is_alive()

    assert ns.kwarg == 0


# @nose.SkipTest  # to help debugging ( FIXME : how to programmatically start only one test - maybe in fixture - ? )
@pytest.mark.timeout(5)
def test_node_as_context_manager():
    """Checks that a node can be used as a context manager"""
    with pyzmp.Node() as n1:  # this will __init__ and __enter__
        assert n1.is_alive()
    assert not n1.is_alive()


# @nose.SkipTest  # to help debugging ( FIXME : how to programmatically start only one test - maybe in fixture - ? )
@pytest.mark.timeout(5)
def test_node_running_as_context_manager():
    """Checks that an already running node can be used as a context manager"""
    n1 = pyzmp.Node()
    n1.start()
    with n1:  # hooking to an already started node
        # This might restart the node (good or bad ?)
        assert n1.is_alive()
    assert not n1.is_alive()


def test_update_rate():
    """
    Testing that the update methods get a correct timedelta
    """
    # TODO : investigate if node multiprocessing plugin would help simplify this
    # playing with list to pass a reference to this
    testing_last_update = [time.time()]
    testing_time_delta = []
    acceptable_timedelta = []

    def testing_update(self, timedelta, last_update, time_delta, ok_timedelta):
        time_delta.append(time.time() - last_update[-1])
        last_update.append(time.time())

        # if the time delta measured in test and the one passed as argument differ
        # too much, one time, test is failed
        if abs(time_delta[-1] - timedelta) > 0.005:
            ok_timedelta.append(False)
        else:
            ok_timedelta.append(True)

        # spin like crazy, loads CPU for a bit, and eventually exits.
        # We re here trying to disturb the update rate
        while True:
            if randint(0, 10000) == 42:
                break

    # hack to dynamically change the update method
    testing_update_onearg = functools.partial(testing_update,
                                        last_update=testing_last_update,
                                        time_delta=testing_time_delta,
                                        ok_timedelta=acceptable_timedelta)

    n1 = pyzmp.Node()
    n1.update = types.MethodType(testing_update_onearg, n1)

    assert not n1.is_alive()

    # Starting the node in the same thread, to be able to test simply by shared memory.
    # TODO : A Node that can choose process or thread run ( on start() instead of init() maybe ? )
    runthread = threading.Thread(target=n1.run)
    runthread.daemon = True  # to kill this when test is finished
    runthread.start()
    # n1.start()

    # sleep here for a while
    time.sleep(10)

    # removing init time only used for delta computation
    testing_last_update.pop(0)
    # Check time vars modified by update
    for i in range(0, len(testing_last_update)):
        print("update : {u} | delta: {d} | accept : {a}".format(
            u=testing_last_update[i],
            d=testing_time_delta[i],
            a=acceptable_timedelta[i])
        )

        assert acceptable_timedelta[i]





### TODO : more testing in case of crash in process, exception, signal, etc.

if __name__ == '__main__':

    import nose
    nose.runmodule()



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from collections import namedtuple


"""
Protocol allowing dynamic specification of message format
"""
# CAREFUL : topic might not be a complete self sufficient concept ( like service )
# TODO : study PIPELINE ( and others ) from zmq
# GOAL : find concept that allow like service ( remote version of funciton call ) but with inverted control flow ( callback )

def gen_msg_type(self, name, **kwargs):
    return namedtuple(name, **kwargs)




class Topic(object):

    def __init__(self, name):
        self.name = name
        self.cur_msg = None

    def publish(self, msg):
        self.cur_msg = msg
        return True






# -*- coding: utf-8 -*-
from __future__ import absolute_import

import contextlib
import six
import time
from collections import namedtuple
import zmq
import pickle
#import dill as pickle
import inspect

try:
    from six import reraise
except ImportError:  # if six is not present we will not reraise the remote exception
    reraise = None

try:
    from tblib import Traceback
except ImportError:  # if tblib is not present, we will not be able to forward the traceback
    Traceback = None

from .message import ServiceRequest, ServiceResponse, ServiceResponse_dictparse, ServiceException, ServiceException_dictparse

from .master import manager
from .exceptions import UnknownResponseTypeException

# Lock is definitely needed ( not implemented in proxy objects, unless the object itself already has it, like Queue )
services_lock = manager.Lock()
services = manager.dict()


@contextlib.contextmanager
def service_provider_cm(node_name, svc_address, node_providers):
    # advertising services
    services_lock.acquire()
    for svc_name, svc_endpoint in six.iteritems(node_providers):
        # print('-> Providing {0} with {1}'.format(svc_name, svc_endpoint))
        # needs reassigning to propagate update to manager
        services[svc_name] = (services[svc_name] if svc_name in services else []) + [(node_name, svc_address)]
    services_lock.release()

    yield

    # concealing services
    services_lock.acquire()
    for svc_name, svc_endpoint in six.iteritems(node_providers):
        # print('-> Unproviding {0}'.format(svc_name))
        services[svc_name] = [(n, a) for (n, a) in services[svc_name] if n != node_name]
    services_lock.release()


class ServiceCallTimeout(Exception):
    pass


# TODO : make this pickleable so we can move it around ( and dynamically rebuild the socket connection )
class Service(object):

    @staticmethod
    # TODO : optionally adds more element from the signature ( num args, etc. )
    def discover(name, timeout=None, minimum_providers=1):
        """
        discovers a service. If timeout is specified, waits for at least minimum_providers service instance to be available.
        Note : we do not want to make the discovery block undefinitely since we never know for sure if a service is running or not
        TODO : improve with future...
        :param name: name of the service
        :param timeout: maximum number of seconds the discover can wait for a discovery matching requirements. if None, doesn't wait.
        :param minimum_providers the number of provider we need to reach before discover() returns if timeout enabled
        :return: a Service object, containing the list of providers. if the minimum number cannot be reached, still returns what is available.
        """
        start = time.time()
        endtime = timeout if timeout else 0

        while True:
            timed_out = time.time() - start > endtime
            if name in services and isinstance(services[name], list):
                if len(services[name]) >= minimum_providers or timed_out:
                    providers = services[name]
                    if providers:
                        return Service(name, providers)
                    else:
                        return None

            if timed_out:
                break
            # else we keep looping after a short sleep ( to allow time to refresh services list )
            time.sleep(0.2)  # sleep
        return None

    def __init__(self, name, providers=None):
        self.name = name
        self.providers = providers
        # TODO : make a provide just a list of node names, and have connection URLs somewhere else...

    # TODO : implement async_call ( and return future )
    def call(self, args=None, kwargs=None, node=None, send_timeout=1000, recv_timeout=5000, zmq_ctx=None):
        """
        Calls a service on a node with req as arguments. if node is None, a node is chosen by zmq.
        if zmq_ctx is passed, it will use the existing context
        Uses a REQ socket. Ref : http://api.zeromq.org/2-1:zmq-socket
        :param node : the node name

        """

        context = zmq_ctx or zmq.Context()
        assert isinstance(context, zmq.Context)

        args = args or ()
        assert isinstance(args, tuple)
        kwargs = kwargs or {}
        assert isinstance(kwargs, dict)

        socket = context.socket(zmq.REQ)

        # connect to all addresses ( optionally matching node name )
        for n, a in [(n, a) for (n, a) in self.providers if (not node or n == node)]:
            socket.connect(a)

        # build message
        fullreq = ServiceRequest(service=self.name, args=pickle.dumps(args), kwargs=pickle.dumps(kwargs))

        poller = zmq.Poller()
        poller.register(socket)  # POLLIN for recv, POLLOUT for send

        evts = dict(poller.poll(send_timeout))
        if socket in evts and evts[socket] == zmq.POLLOUT:
            socket.send(fullreq.serialize())
            # TODO : find a way to get rid fo these timeouts when debugging
            # TODO : when timeout Exception should occur ( not returning None )
            evts = dict(poller.poll(recv_timeout))  # blocking until answer
            if socket in evts and evts[socket] == zmq.POLLIN:
                resp = socket.recv()
                fullresp = ServiceResponse_dictparse(resp)

                if fullresp.has_field('response'):
                    return pickle.loads(fullresp.response)
                elif fullresp.has_field('exception'):
                    svcexc = fullresp.exception  # This has already been parsed by ServiceResponse_dictparse
                    tb = pickle.loads(svcexc.traceback)
                    if Traceback and isinstance(tb, Traceback):
                        reraise(pickle.loads(svcexc.exc_type), pickle.loads(svcexc.exc_value), tb.as_traceback())
                    else:  # traceback not usable
                        reraise(pickle.loads(svcexc.exc_type), pickle.loads(svcexc.exc_value), None)
                else:
                    raise UnknownResponseTypeException("Unknown Response Type {0}".format(type(fullresp)))
            else:
                raise ServiceCallTimeout("Did not receive response through ZMQ socket.")
        else:
            raise ServiceCallTimeout("Can not send request through ZMQ socket.")

        # TODO : exception if service disappeared in the mean time...

# convenience
discover = Service.discover



# -*- coding: utf-8 -*-
# This python package is implementing a very simple multiprocess framework
# The point of it is to be able to fully tests the multiprocess behavior,
#     in pure python, without having to run a ROS system.
from __future__ import absolute_import

import multiprocessing

#https://pymotw.com/2/multiprocessing/communication.html
#https://docs.python.org/2/library/multiprocessing.html#proxy-objects



"""
Keeps a list of topics, services, params, nodes available on the system
Very very simple implementation for now, as it is not our main focus to build yet another full fledged distributed multiprocessing framework
"""
manager = multiprocessing.Manager()





# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function



class UnknownServiceException(Exception):
    """
    Raised by a Service call if the destination node doesnt provide this service
    """
    pass


class UnknownRequestTypeException(Exception):
    """
    Raised by a Service call if the destination node doesnt recognize the request type
    """
    pass


class UnknownResponseTypeException(Exception):
    """
    Raised by a Service call if the service doesnt recognise the response type
    """
    pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

import pickle
import pyzmp.message

import pytest
# http://pytest.org/latest/contents.html
# https://github.com/ionelmc/pytest-benchmark

# TODO : PYPY
# http://pypy.org/

# Careful : we dont support all fancy methods of the protocol buffer.
# We restrict ourselves to the methods that we are using in pyzmp.


# Generic initialization tests
def initialize_servicerequest(req):
    assert req.service == "testservice"
    assert pickle.loads(req.args)[0] == "testarg"
    assert pickle.loads(req.kwargs)["testkwarg"] == "test"
    assert req.initialized()


def initialize_serviceresponse(resp):
    assert resp.service == "testservice"
    assert resp.has_field('response') == True
    assert resp.has_field('exception') == False
    assert pickle.loads(resp.response) == "testresponse"
    assert resp.initialized()


def initialize_serviceexception(resp):
    assert resp.exc_type == "testexception"
    assert pickle.loads(resp.exc_value) == "testexceptionvalue"
    assert pickle.loads(resp.traceback) == "testtraceback"
    assert resp.initialized()


def initialize_serviceresponseexception(resp):
    assert resp.service == "testservice"
    assert resp.has_field('response') == False
    assert resp.has_field('exception') == True
    assert resp.exception.exc_type == "testexception"
    assert pickle.loads(resp.exception.exc_value) == "testexceptionvalue"
    assert pickle.loads(resp.exception.traceback) == "testtraceback"
    assert resp.initialized()


def default_initialize_servicerequest(req):
    # Although default initialization behaviors are different in protobuf and namedtuple, both should be initialized
    assert req.initialized()


def default_initialize_serviceresponse(resp):
    # Although default initialization behaviors are different in protobuf and namedtuple, both should be initialized
    assert resp.initialized()


def default_initialize_serviceexception(resp):
    # Although default initialization behaviors are different in protobuf and namedtuple, both should be initialized
    assert resp.initialized()


# No point to have a default_initialize_serviceresponseexception here


# Symmetric Serialize / Parse tests
def oneline_symmetric_serialize_parse_servicerequest(req):
    assert req.initialized()
    # Testing oneline implementation that works for both tuples and protobuf
    finalreq = pyzmp.message.ServiceRequest_dictparse(req.serialize())
    assert isinstance(finalreq, pyzmp.message.ServiceRequest)
    assert finalreq.initialized()
    assert finalreq.service == "testservice"
    assert pickle.loads(finalreq.args)[0] == "testarg"
    assert pickle.loads(finalreq.kwargs)["testkwarg"] == "test"


def oneline_symmetric_serialize_parse_serviceresponse(resp):
    assert resp.initialized()
    # Testing oneline implementation that works for both tuples and protobuf
    finalresp = pyzmp.message.ServiceResponse_dictparse(resp.serialize())
    assert isinstance(finalresp, pyzmp.message.ServiceResponse)
    assert finalresp.initialized()
    assert finalresp.service == "testservice"
    assert finalresp.has_field('response') == True
    assert finalresp.has_field('exception') == False
    assert pickle.loads(finalresp.response) == "testresponse"


def oneline_symmetric_serialize_parse_serviceexception(resp):
    assert resp.initialized()
    # Testing oneline implementation that works for both tuples and protobuf
    finalresp = pyzmp.message.ServiceException_dictparse(resp.serialize())
    assert isinstance(finalresp, pyzmp.message.ServiceException)
    assert finalresp.initialized()
    assert finalresp.exc_type == "testexception"
    assert pickle.loads(finalresp.exc_value) == "testexceptionvalue"
    assert pickle.loads(finalresp.traceback) == "testtraceback"


def oneline_symmetric_serialize_parse_serviceresponseexception(resp):
    assert resp.initialized()
    # Testing oneline implementation that works for both tuples and protobuf
    finalresp = pyzmp.message.ServiceResponse_dictparse(resp.serialize())
    assert isinstance(finalresp, pyzmp.message.ServiceResponse)
    assert finalresp.initialized()
    assert finalresp.service == "testservice"
    assert finalresp.has_field('response') ==  False
    assert finalresp.has_field('exception') ==  True
    assert finalresp.exception.exc_type ==  "testexception"
    assert pickle.loads(finalresp.exception.exc_value) == "testexceptionvalue"
    assert pickle.loads(finalresp.exception.traceback) == "testtraceback"


if_protobuf = pytest.mark.skipif(not pyzmp.message.protobuf_implementation_enabled,
                    reason="Protobuf implementation not supported !")

# PROTOBUF Implementation test - default
class TestMessageProtobuf(object):
    def setup_method(self, method):
        # Checking protobuf implementation
        pass

    def teardown_method(self, method):
        # Setting back default implementation
        pass

    @if_protobuf
    def test_initialize_servicerequest_protobuf(self):
        # Test Initialization
        req = pyzmp.message.ServiceRequest(
            service="testservice",
            args=pickle.dumps(("testarg",)),
            kwargs=pickle.dumps({'testkwarg': 'test'}),
        )
        # Check we have desired implementation
        assert isinstance(req, pyzmp.message.ServiceRequestImpl)
        # Check it is an instance of Dynamic Functional Facade
        assert isinstance(req, pyzmp.message.ServiceRequest)
        # run actual test
        initialize_servicerequest(req)

    @if_protobuf
    def test_initialize_serviceresponse_protobuf(self):
        #Test Initialization
        resp = pyzmp.message.ServiceResponse(
            service="testservice",
            response=pickle.dumps("testresponse"),
        )
        # Check we have desired implementation
        assert isinstance(resp, pyzmp.message.ServiceResponseImpl)
        # Check it is an instance of Dynamic Functional Facade
        assert isinstance(resp, pyzmp.message.ServiceResponse)
        # run actual test
        initialize_serviceresponse(resp)

    @if_protobuf
    def test_initialize_serviceexception_protobuf(self):
        #Test Initialization
        resp = pyzmp.message.ServiceException(
            exc_type="testexception",
            exc_value=pickle.dumps("testexceptionvalue"),
            traceback=pickle.dumps("testtraceback"),
        )
        # Check we have desired implementation
        assert isinstance(resp, pyzmp.message.ServiceExceptionImpl)
        # Check it is an instance of Dynamic Functional Facade
        assert isinstance(resp, pyzmp.message.ServiceException)
        # run actual test
        initialize_serviceexception(resp)

    @if_protobuf
    def test_initialize_serviceresponseexception_protobuf(self):
        #Test Initialization
        resp = pyzmp.message.ServiceResponse(
            service="testservice",
            exception=pyzmp.message.ServiceException(
                exc_type="testexception",
                exc_value=pickle.dumps("testexceptionvalue"),
                traceback=pickle.dumps("testtraceback"),
            )
        )
        # Check we have desired implementation
        assert isinstance(resp, pyzmp.message.ServiceResponseImpl)
        # Check it is an instance of Dynamic Functional Facade
        assert isinstance(resp, pyzmp.message.ServiceResponse)
        # run actual test
        initialize_serviceresponseexception(resp)

    @if_protobuf
    def test_default_initialize_servicerequest_protobuf(self):
        # Test Initialization
        req = pyzmp.message.ServiceRequest()
        # Check we have desired implementation
        assert isinstance(req, pyzmp.message.ServiceRequestImpl)
        # Check it is an instance of Dynamic Functional Facade
        assert isinstance(req, pyzmp.message.ServiceRequest)
        # run actual test
        default_initialize_servicerequest(req)

    @if_protobuf
    def test_default_initialize_serviceresponse_protobuf(self):
        # Test Initialization
        resp = pyzmp.message.ServiceResponse()
        # Check we have desired implementation
        assert isinstance(resp, pyzmp.message.ServiceResponseImpl)
        # Check it is an instance of Dynamic Functional Facade
        assert isinstance(resp, pyzmp.message.ServiceResponse)
        # run actual test
        default_initialize_serviceresponse(resp)

    @if_protobuf
    def test_default_initialize_serviceexception_protobuf(self):
        # Test Initialization
        exc = pyzmp.message.ServiceException()
        # Check we have desired implementation
        assert isinstance(exc, pyzmp.message.ServiceExceptionImpl)
        # Check it is an instance of Dynamic Functional Facade
        assert isinstance(exc, pyzmp.message.ServiceException)
        # run actual test
        default_initialize_serviceexception(exc)


    # No point to have a test_default_initialize_serviceresponseexception_protobuf here

    @if_protobuf
    def test_symmetric_serialize_parse_servicerequest_protobuf(self):
        # Test Initialization
        req = pyzmp.message.ServiceRequest(
            service="testservice",
            args=pickle.dumps(("testarg",)),
            kwargs=pickle.dumps({'testkwarg': 'test'}),
        )
        # Check we have desired implementation
        assert isinstance(req, pyzmp.message.ServiceRequestImpl)
        # Check it is an instance of Dynamic Functional Facade
        assert isinstance(req, pyzmp.message.ServiceRequest)

        assert req.initialized()

        # assert parse modifies the current message
        parsedreq = pyzmp.message.ServiceRequest()
        parsedreq._parse(req.serialize())
        assert isinstance(parsedreq, pyzmp.message.ServiceRequest)
        assert parsedreq.initialized()
        assert parsedreq.service ==  "testservice"
        assert pickle.loads(parsedreq.args)[0] ==  "testarg"
        assert pickle.loads(parsedreq.kwargs)["testkwarg"] ==  "test"

        # assert parse returns the modified list
        parsedreq_oneline = pyzmp.message.ServiceRequest()._parse(req.serialize())
        assert isinstance(parsedreq_oneline, pyzmp.message.ServiceRequest)
        assert parsedreq_oneline.initialized()
        assert parsedreq_oneline.service ==  "testservice"
        assert pickle.loads(parsedreq_oneline.args)[0] ==  "testarg"
        assert pickle.loads(parsedreq_oneline.kwargs)["testkwarg"] ==  "test"

        # run actual tuple-compatible online test
        oneline_symmetric_serialize_parse_servicerequest(req)

    @if_protobuf
    def test_symmetric_serialize_parse_serviceresponse_protobuf(self):
        #Test Initialization
        resp = pyzmp.message.ServiceResponse(
            service="testservice",
            response=pickle.dumps("testresponse")
        )
        # Check we have desired implementation
        assert isinstance(resp, pyzmp.message.ServiceResponseImpl)
        # Check it is an instance of Dynamic Functional Facade
        assert isinstance(resp, pyzmp.message.ServiceResponse)

        assert resp.initialized()

        # assert parse modifies the current message
        parsedresp = pyzmp.message.ServiceResponse()
        parsedresp._parse(resp.serialize())
        assert isinstance(parsedresp, pyzmp.message.ServiceResponse)
        assert parsedresp.initialized()
        assert parsedresp.service ==  "testservice"
        assert pickle.loads(parsedresp.response) ==  "testresponse"

        # assert parse returns the modified list
        parsedresp_oneline = pyzmp.message.ServiceResponse()._parse(resp.serialize())
        assert isinstance(parsedresp_oneline, pyzmp.message.ServiceResponse)
        assert parsedresp_oneline.initialized()
        assert parsedresp_oneline.service ==  "testservice"
        assert pickle.loads(parsedresp_oneline.response) ==  "testresponse"

        # run actual tuple-compatible online test
        oneline_symmetric_serialize_parse_serviceresponse(resp)

    @if_protobuf
    def test_symmetric_serialize_parse_serviceexception_protobuf(self):
        #Test Initialization
        resp = pyzmp.message.ServiceException(
            exc_type="testexception",
            exc_value=pickle.dumps("testexceptionvalue"),
            traceback=pickle.dumps("testtraceback"),
        )
        # Check we have desired implementation
        assert isinstance(resp, pyzmp.message.ServiceExceptionImpl)
        # Check it is an instance of Dynamic Functional Facade
        assert isinstance(resp, pyzmp.message.ServiceException)

        assert resp.initialized()

        # assert parse modifies the current message
        parsedresp = pyzmp.message.ServiceException()
        parsedresp._parse(resp.serialize())
        assert isinstance(parsedresp, pyzmp.message.ServiceException)
        assert parsedresp.initialized()
        assert parsedresp.exc_type ==  "testexception"
        assert pickle.loads(parsedresp.exc_value) ==  "testexceptionvalue"
        assert pickle.loads(parsedresp.traceback) ==  "testtraceback"

        # assert parse returns the modified list
        parsedresp_oneline = pyzmp.message.ServiceException()._parse(resp.serialize())
        assert isinstance(parsedresp_oneline, pyzmp.message.ServiceException)
        assert parsedresp_oneline.initialized()
        assert parsedresp_oneline.exc_type ==  "testexception"
        assert pickle.loads(parsedresp_oneline.exc_value) ==  "testexceptionvalue"
        assert pickle.loads(parsedresp_oneline.traceback) ==  "testtraceback"

        # run actual tuple-compatible online test
        oneline_symmetric_serialize_parse_serviceexception(resp)

    @if_protobuf
    def test_symmetric_serialize_parse_serviceresponseexception_protobuf(self):
        #Test Initialization
        resp = pyzmp.message.ServiceResponse(
            service="testservice",
            exception=pyzmp.message.ServiceException(
                exc_type="testexception",
                exc_value=pickle.dumps("testexceptionvalue"),
                traceback=pickle.dumps("testtraceback"),
            )
        )
        # Check we have desired implementation
        assert isinstance(resp, pyzmp.message.ServiceResponseImpl)
        # Check it is an instance of Dynamic Functional Facade
        assert isinstance(resp, pyzmp.message.ServiceResponse)

        assert resp.initialized()

        # assert parse modifies the current message
        parsedresp = pyzmp.message.ServiceResponse()
        parsedresp._parse(resp.serialize())
        assert isinstance(parsedresp, pyzmp.message.ServiceResponse)
        assert parsedresp.initialized()
        assert parsedresp.service ==  "testservice"
        assert parsedresp.HasField('response') ==  False
        assert parsedresp.HasField('exception') ==  True
        assert parsedresp.exception.exc_type ==  "testexception"
        assert pickle.loads(parsedresp.exception.exc_value) ==  "testexceptionvalue"
        assert pickle.loads(parsedresp.exception.traceback) ==  "testtraceback"

        # assert parse returns the modified list
        parsedresp_oneline = pyzmp.message.ServiceResponse()._parse(resp.serialize())
        assert isinstance(parsedresp_oneline, pyzmp.message.ServiceResponse)
        assert parsedresp_oneline.initialized()
        assert parsedresp_oneline.service ==  "testservice"
        assert parsedresp_oneline.HasField('response') ==  False
        assert parsedresp_oneline.HasField('exception') ==  True
        assert parsedresp_oneline.exception.exc_type ==  "testexception"
        assert pickle.loads(parsedresp_oneline.exception.exc_value) ==  "testexceptionvalue"
        assert pickle.loads(parsedresp_oneline.exception.traceback) ==  "testtraceback"

        # run actual tuple-compatible online test
        oneline_symmetric_serialize_parse_serviceresponseexception(resp)


class TestMessageTupleFallback(object):
    def setup_method(self, method):
        # Forcing tuple implementation
        pyzmp.message.force_namedtuple_implementation()

    def teardown_method(self, method):
        # Setting back default implementation
        pyzmp.message.force_protobuf_implementation()

    def test_initialize_servicerequest_namedtuple(self):
        #Test Initialization
        req = pyzmp.message.ServiceRequest(
            service="testservice",
            args=pickle.dumps(("testarg",)),
            kwargs=pickle.dumps({'testkwarg': 'test'}),
        )
        # Check we have desired implementation
        assert isinstance(req, pyzmp.message.ServiceRequestNTImpl)
        # Check it is an instance of Dynamic Functional Facade
        assert isinstance(req, pyzmp.message.ServiceRequest)
        # run actual test
        initialize_servicerequest(req)

    def test_initialize_serviceresponse_namedtuple(self):
        #Test Initialization
        resp = pyzmp.message.ServiceResponse(
            service="testservice",
            response=pickle.dumps("testresponse"),
            exception=None,
        )
        # Check we have desired implementation
        assert isinstance(resp, pyzmp.message.ServiceResponseNTImpl)
        # Check it is an instance of Dynamic Functional Facade
        assert isinstance(resp, pyzmp.message.ServiceResponse)
        # run actual test
        initialize_serviceresponse(resp)

    def test_initialize_serviceexception_namedtuple(self):
        #Test Initialization
        resp = pyzmp.message.ServiceException(
            exc_type="testexception",
            exc_value=pickle.dumps("testexceptionvalue"),
            traceback=pickle.dumps("testtraceback"),
        )
        # Check we have desired implementation
        assert isinstance(resp, pyzmp.message.ServiceExceptionNTImpl)
        # Check it is an instance of Dynamic Functional Facade
        assert isinstance(resp, pyzmp.message.ServiceException)
        # run actual test
        initialize_serviceexception(resp)

    def test_initialize_serviceresponseexception_namedtuple(self):
        #Test Initialization
        resp = pyzmp.message.ServiceResponse(
            service="testservice",
            response=None,
            exception=pyzmp.message.ServiceException(
                exc_type="testexception",
                exc_value=pickle.dumps("testexceptionvalue"),
                traceback=pickle.dumps("testtraceback"),
            )
        )
        # Check we have desired implementation
        assert isinstance(resp, pyzmp.message.ServiceResponseNTImpl)
        # Check it is an instance of Dynamic Functional Facade
        assert isinstance(resp, pyzmp.message.ServiceResponse)
        # run actual test
        initialize_serviceresponseexception(resp)

    def test_default_initialize_servicerequest_namedtuple(self):
        # Test Initialization
        req = pyzmp.message.ServiceRequest()
        # Check we have desired implementation
        assert isinstance(req, pyzmp.message.ServiceRequestNTImpl)
        # Check it is an instance of Dynamic Functional Facade
        assert isinstance(req, pyzmp.message.ServiceRequest)
        # run actual test
        default_initialize_servicerequest(req)

    def test_default_initialize_serviceresponse_namedtuple(self):
        # Test Initialization
        resp = pyzmp.message.ServiceResponse()
        # Check we have desired implementation
        assert isinstance(resp, pyzmp.message.ServiceResponseNTImpl)
        # Check it is an instance of Dynamic Functional Facade
        assert isinstance(resp, pyzmp.message.ServiceResponse)
        # run actual test
        default_initialize_serviceresponse(resp)

    def test_default_initialize_serviceexception_namedtuple(self):
        # Test Initialization
        resp = pyzmp.message.ServiceException()
        # Check we have desired implementation
        assert isinstance(resp, pyzmp.message.ServiceExceptionNTImpl)
        # Check it is an instance of Dynamic Functional Facade
        assert isinstance(resp, pyzmp.message.ServiceException)
        # run actual test
        default_initialize_serviceexception(resp)

    def test_symmetric_serialize_parse_servicerequest_namedtuple(self):
        # Test Initialization
        req = pyzmp.message.ServiceRequest(
            service="testservice",
            args=pickle.dumps(("testarg",)),
            kwargs=pickle.dumps({'testkwarg': 'test'}),
        )
        # Check we have desired implementation
        assert isinstance(req, pyzmp.message.ServiceRequestNTImpl)
        # Check it is an instance of Dynamic Functional Facade
        assert isinstance(req, pyzmp.message.ServiceRequest)

        assert req.initialized()

        # run actual protobuf-compatible online test
        oneline_symmetric_serialize_parse_servicerequest(req)

    def test_symmetric_serialize_parse_serviceresponse_namedtuple(self):
        #Test Initialization
        resp = pyzmp.message.ServiceResponse(
            service="testservice",
            response=pickle.dumps("testresponse"),
            exception=None,
        )
        # Check we have desired implementation
        assert isinstance(resp, pyzmp.message.ServiceResponseNTImpl)
        # Check it is an instance of Dynamic Functional Facade
        assert isinstance(resp, pyzmp.message.ServiceResponse)

        assert resp.initialized()

        # run actual protobuf-compatible online test
        oneline_symmetric_serialize_parse_serviceresponse(resp)


    def test_symmetric_serialize_parse_serviceexception_namedtuple(self):
        #Test Initialization
        resp = pyzmp.message.ServiceException(
            exc_type="testexception",
            exc_value=pickle.dumps("testexceptionvalue"),
            traceback=pickle.dumps("testtraceback"),
        )
        # Check we have desired implementation
        assert isinstance(resp, pyzmp.message.ServiceExceptionNTImpl)
        # Check it is an instance of Dynamic Functional Facade
        assert isinstance(resp, pyzmp.message.ServiceException)

        assert resp.initialized()

        # run actual protobuf-compatible online test
        oneline_symmetric_serialize_parse_serviceexception(resp)

    def test_symmetric_serialize_parse_serviceresponseexception_namedtuple(self):
        #Test Initialization
        resp = pyzmp.message.ServiceResponse(
            service="testservice",
            response=None,
            exception=pyzmp.message.ServiceException(
                exc_type="testexception",
                exc_value=pickle.dumps("testexceptionvalue"),
                traceback=pickle.dumps("testtraceback"),
            ),
        )
        # Check we have desired implementation
        assert isinstance(resp, pyzmp.message.ServiceResponseNTImpl)
        # Check it is an instance of Dynamic Functional Facade
        assert isinstance(resp, pyzmp.message.ServiceResponse)

        assert resp.initialized()

        # run actual protobuf-compatible online test
        oneline_symmetric_serialize_parse_serviceresponseexception(resp)




# -*- coding: utf-8 -*-
# This python package is implementing a very simple multiprocess framework
# The point of it is to be able to fully tests the multiprocess behavior,
#     in pure python, without having to run a ROS system.
from __future__ import absolute_import
from __future__ import print_function

import sys
import tempfile
import multiprocessing, multiprocessing.reduction
import types
import zmq
import socket
import logging
import pickle
import contextlib
#import dill as pickle

# allowing pickling of exceptions to transfer it
from collections import namedtuple

import time

try:
    from tblib.decorators import Traceback
    # TODO : potential candidates for pickle + tblib replacement for easier serialization
    # TODO : - https://github.com/uqfoundation/dill
    # TODO : - OR https://github.com/cloudpipe/cloudpickle
    # TODO : - OR https://github.com/irmen/Serpent ?
    # TODO : - OR https://github.com/esnme/ultrajson ?
    # TODO : - OR something else ?
except ImportError:
    Traceback = None

try:
    import setproctitle
except ImportError:
    setproctitle = None

### IMPORTANT : COMPOSITION -> A SET OF NODE SHOULD ALSO 'BE' A NODE ###
### IMPORTANT : IDENTITY
### Category Theory https://en.wikipedia.org/wiki/Category_theory

### Data Flow with topic :
#
# object : message
# arrow : topic_listener
# binary op - associativity : listener l1, l2, l3 <= (l1 . l2) .l3 == l1. (l2. l3)
# binary op - identity : noop on listener => msg transfer as is
# -> Expressed in programming language (functional programming follows category theory) :
# msg3 = topic2_listener(topic1_listener(msg1))

# -> Expressed in graph :
# msg3 <--topic2_listener-- msg2 <--topic1_listener-- msg1

### RPC with services :
#
# object : (req, resp)
# arrow : service_call
# binary op - associativity : service s1, s2, s3 <= (s1 . s2) .s3 == s1. (s2. s3)
# binary op - identity : noop on service => (req, resp) transfer as is ( two ways comm )
# -> Expressed in programming language (functional programming follows category theory) :
#  (req3, resp3) = service2_call(service1_call((req1, resp1)))

# -> Expressed in graph :
# msg1 --service1_call--> msg2 --service2_call--> msg3

###### higher level for execution graph to be represented by a category ######

# TODO : CAREFUL topic is probably not a complete concept in itself
# => often we need to get back to pub / sub connections rather than dealing with "topics"
# We need the reverse flow of "service call" which is a distributed generalisation of function call
# We need the distributed generalisation of the callback function ( async calls and futures )

### Service is a first class citizen. node is abstracted in that perspective : implementation requires some discovery mechanism.
# object : service
# arrow : call
# binary op - associativity : call c1, c2, c3 <= (c1 . c2) .c3 == c1 . (c2 . c3)
# binary op - identity : ??? TODO
# -> Expressed in programming language (functional programming follows category theory) :
#  svc1 = lambda x: return ( lambda y: return svc3(y) )( x )  <= currying/partial => svc1 = lambda x, y : return svc23(x,y)

# -> Expressed in graph :
# svc1 --call--> svc2 --call--> svc3

### Node is a first class citizen
# object : node
# arrow : topic_listener / callback
# binary op - associativity : callback cb1. cb2. cb3 <= (cb1 . cb2) . cb3 == cb1 . (cb2 . cb3)
# binary op - identity : ??? TODO
# -> Expressed in programming language (functional programming follows category theory) :
#  node1.cb = lambda x: return ( lambda y: return node3.cb(y) )(x) <= currying/partial => node1.cb = lambda x, y : return node23(x,y)

# -> Expressed in graph :
# node3 <--topic_cb-- node2 <--topic_cb-- node1

from .master import manager
from .exceptions import UnknownServiceException, UnknownRequestTypeException
from .message import ServiceRequest, ServiceRequest_dictparse, ServiceResponse, ServiceException
from .service import service_provider_cm
#from .service import RequestMsg, ResponseMsg, ErrorMsg  # only to access message types

current_node = multiprocessing.current_process

# Lock is definitely needed ( not implemented in proxy objects, unless the object itself already has it, like Queue )
nodes_lock = manager.Lock()
nodes = manager.dict()


@contextlib.contextmanager
def dummy_cm():
    yield None


@contextlib.contextmanager
def node_cm(node_name, svc_address):
    # advertise itself
    nodes_lock.acquire()
    nodes[node_name] = {'service_conn': svc_address}
    nodes_lock.release()

    yield

    # concealing itself
    nodes_lock.acquire()
    nodes[node_name] = {}
    nodes_lock.release()


# TODO : Nodelet ( thread, with fast intraprocess zmq comm - entity system design /vs/threadpool ?)
# CAREFUL here : multiprocessing documentation specifies that a process object can be started only once...
class Node(object):

    EndPoint = namedtuple("EndPoint", "self func")

    # TODO : allow just passing target to be able to make a Node from a simple function, and also via decorator...
    def __init__(self, name='pyzmp_node', socket_bind=None, context_manager=None, target=None, args=None, kwargs=None):
        """
        Initializes a ZMP Node (Restartable Python Process communicating via ZMQ)
        :param name: Name of the node
        :param socket_bind: the string describing how to bind the ZMQ socket ( IPC, TCP, etc. )
        :param context_manager: a context_manager to be used with run (in a with statement)
        :return:
        """
        # TODO check name unicity
        # using process as delegate
        self._pargs = {
            'name': name,
            'args': args or (),
            'kwargs': kwargs or {},
            'target': self.run,  # Careful : our run() is not the same as the one for Process
        }
        # Careful : our own target is not the same as the one for Process
        self._target = target or self.update  # we expect the user to overload update in child class.

        #: the actual process instance. lazy creation on start() call only.
        self._process = None

        #: whether or not the node name should be set as the actual process title
        #: replacing the string duplicated from the python interpreter run
        self.new_title = True

        self.context_manager = context_manager or dummy_cm  # TODO: extend to list if possible ( available for python >3.1 only )
        self.exit = multiprocessing.Event()
        self.started = multiprocessing.Event()
        self.listeners = {}
        self._providers = {}
        self.tmpdir = tempfile.mkdtemp(prefix='zmp-' + self.name + '-')
        # if no socket is specified the services of this node will be available only through IPC
        self._svc_address = socket_bind if socket_bind else 'ipc://' + self.tmpdir + '/services.pipe'

    def __enter__(self):
        # __enter__ is called only if we pass this instance to with statement ( after __init__ )
        # start only if needed (so that we can hook up a context manager to a running node) :
        if not self.is_alive():
            self.start()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        # make sure we cleanup when we exit
        self.shutdown()

    def has_started(self):
        """
        :return: True if the node has started (update() called at least once). Might still be alive, or not...
        """
        return self.started.is_set()

    ### Process API delegation ###
    def is_alive(self):
        if self and self._process:
            return self._process.is_alive()

    def join(self, timeout=None):
        if not self._process:
            # blocking on started event before blocking on join
            self.started.wait(timeout=timeout)
        return self._process.join(timeout=timeout)

    @property
    def name(self):
        if self and self._process:
            return self._process.name
        else:
            return self._pargs.get('name', "ZMPNode")

    @name.setter
    def name(self, name):
        if self and self._process:
            self._process.name = name
            # only reset the name arg if it was accepted by the setter
            self._pargs.set('name', self._process.name)
        else:
            # TODO : maybe we should be a bit more strict here ?
            self._pargs.set('name', name)

    @property
    def daemon(self):
        """
        Return whether process is a daemon
        :return:
        """
        if self._process:
            return self._process.daemon
        else:
            return self._pargs.get('daemonic', False)

    @daemon.setter
    def daemon(self, daemonic):
        """
        Set whether process is a daemon
        :param daemonic:
        :return:
        """
        if self._process:
            self._process.daemonic = daemonic
        else:
            self._pargs['daemonic'] = daemonic

    @property
    def authkey(self):
        return self._process.authkey

    @authkey.setter
    def authkey(self, authkey):
        """
        Set authorization key of process
        """
        self._process.authkey = authkey

    @property
    def exitcode(self):
        """
        Return exit code of process or `None` if it has yet to stop
        """
        if self._process:
            return self._process.exitcode
        else:
            return None

    @property
    def ident(self):
        """
        Return identifier (PID) of process or `None` if it has yet to start
        """
        if self._process:
            return self._process.ident
        else:
            return None

    def __repr__(self):
        # TODO : improve this
        return self._process.__repr__()

    def start(self, timeout=None):
        """
        Start child process
        :param timeout: the maximum time to wait for child process to report it has actually started.
        None waits until the update has been called at least once.
        """

        # we lazily create our process delegate (with same arguments)
        if self.daemon:
            daemonic = True
        else:
            daemonic = False

        pargs = self._pargs.copy()
        pargs.pop('daemonic', None)

        self._process = multiprocessing.Process(**pargs)

        self._process.daemon = daemonic

        if self.is_alive():
            # if already started, we shutdown and join before restarting
            # not timeout will bock here (default join behavior).
            # otherwise we simply use the same timeout.
            self.shutdown(join=True, timeout=timeout)  # TODO : only restart if no error (check exitcode)
            self.start(timeout=timeout)  # recursive to try again if needed
        else:
            self._process.start()

        # timeout None means we want to wait and ensure it has started
        # deterministic behavior, like is_alive() from multiprocess.Process is always true after start()
        return self.started.wait(timeout=timeout)  # blocks until we know true or false
        # TODO : here we should probably return the zmp url as interface to connect to the node...
        # TODO: futures and ThreadPoolExecutor (so we dont need to manage the pool ourselves)

    # TODO : Implement a way to redirect stdout/stderr, or even forward to parent ?
    # cf http://ryanjoneil.github.io/posts/2014-02-14-capturing-stdout-in-a-python-child-process.html

    def terminate(self):
        """Forcefully terminates the underlying process (using SIGTERM)"""
        return self._process.terminate()
        # TODO : maybe redirect to shutdown here to avoid child process leaks ?

    ### Node specific API ###
    # TODO : find a way to separate process management and service provider API

    def provides(self, svc_callback, service_name=None):
        service_name = service_name or svc_callback.__name__
        # we store an endpoint ( bound method or unbound function )
        self._providers[service_name] = Node.EndPoint(
            self=getattr(svc_callback, '__self__', None),
            func=getattr(svc_callback, '__func__', svc_callback),
        )

    def withholds(self, service_name):
        service_name = getattr(service_name, '__name__', service_name)
        # we store an endpoint ( bound method or unbound function )
        self._providers.pop(service_name)

    # TODO : shortcut to discover/build only services provided by this node ?

    # Careful : this is NOT the same usage as "run()" from Process :
    # it is called inside a loop that it does not directly control...
    # TOOD : think about it and improve (Entity System integration ? Pool + Futures integration ?)
    def update(self, *args, **kwargs):
        """
        Runs at every update cycle in the node process/thread.
        Usually you want to override this method to extend the behavior of the node in your implementation
        :return: integer as exitcode to stop the node, or None to keep looping...
        """
        # TODO : Check which way is better (can also be used to run external process, other functions, like Process)
        return None  # we keep looping by default (need to deal with services here...)

    def shutdown(self, join=True, timeout=None):
        """
        Clean shutdown of the node.
        :param join: optionally wait for the process to end (default : True)
        :return: None
        """
        if self.is_alive():  # check if process started
            print("Shutdown initiated")
            self.exit.set()
            if join:
                self.join(timeout=timeout)
                # TODO : timeout before forcing terminate (SIGTERM)

        exitcode = self._process.exitcode if self._process else None  # we return None if the process was never started
        return exitcode

    def run(self, *args, **kwargs):
        """
        The Node main method, running in a child process (similar to Process.run() but also accepts args)
        A children class can override this method, but it needs to call super().run(*args, **kwargs)
        for the node to start properly and call update() as expected.
        :param args: arguments to pass to update()
        :param kwargs: keyword arguments to pass to update()
        :return: last exitcode returned by update()
        """
        # TODO : make use of the arguments ? since run is now the target for Process...

        exitstatus = None  # keeping the semantic of multiprocessing.Process : running process has None

        if setproctitle and self.new_title:
            setproctitle.setproctitle("{0}".format(self.name))

        print('[{node}] Node started as [{pid} <= {address}]'.format(node=self.name, pid=self.ident, address=self._svc_address))

        zcontext = zmq.Context()  # check creating context in init ( compatibility with multiple processes )
        # Apparently not needed ? Ref : https://github.com/zeromq/pyzmq/issues/770
        zcontext.setsockopt(socket.SO_REUSEADDR, 1)  # required to make restart easy and avoid debugging traps...
        svc_socket = zcontext.socket(zmq.REP)  # Ref : http://api.zeromq.org/2-1:zmq-socket # TODO : ROUTER instead ?
        svc_socket.bind(self._svc_address,)

        poller = zmq.Poller()
        poller.register(svc_socket, zmq.POLLIN)

        # Initializing all context managers
        with service_provider_cm(
                    self.name, self._svc_address, self._providers
                ), node_cm(self.name, self._svc_address), self.context_manager() as cm:

            # Starting the clock
            start = time.time()

            first_loop = True
            # loop listening to connection
            while not self.exit.is_set():

                # blocking. messages are received ASAP. timeout only determine update/shutdown speed.
                socks = dict(poller.poll(timeout=100))
                if svc_socket in socks and socks[svc_socket] == zmq.POLLIN:
                    req = None
                    try:
                        req_unparsed = svc_socket.recv()
                        req = ServiceRequest_dictparse(req_unparsed)
                        if isinstance(req, ServiceRequest):
                            if req.service and req.service in self._providers.keys():

                                request_args = pickle.loads(req.args) if req.args else ()
                                # add 'self' if providers[req.service] is a bound method.
                                if self._providers[req.service].self:
                                    request_args = (self, ) + request_args
                                request_kwargs = pickle.loads(req.kwargs) if req.kwargs else {}

                                resp = self._providers[req.service].func(*request_args, **request_kwargs)
                                svc_socket.send(ServiceResponse(
                                    service=req.service,
                                    response=pickle.dumps(resp),
                                ).serialize())

                            else:
                                raise UnknownServiceException("Unknown Service {0}".format(req.service))
                        else:  # should not happen : dictparse would fail before reaching here...
                            raise UnknownRequestTypeException("Unknown Request Type {0}".format(type(req.request)))
                    except Exception:  # we transmit back all errors, and keep spinning...
                        exctype, excvalue, tb = sys.exc_info()
                        # trying to make a pickleable traceback
                        try:
                            ftb = Traceback(tb)
                        except TypeError as exc:
                            ftb = "Traceback manipulation error: {exc}. Verify that python-tblib is installed.".format(exc=exc)

                        # sending back that exception with traceback
                        svc_socket.send(ServiceResponse(
                            service=req.service,
                            exception=ServiceException(
                                exc_type=pickle.dumps(exctype),
                                exc_value=pickle.dumps(excvalue),
                                traceback=pickle.dumps(ftb),
                            )
                        ).serialize())

                # time is ticking
                # TODO : move this out of here. this class should require only generic interface to update method.
                now = time.time()
                timedelta = now - start
                start = now

                # replacing the original Process.run() call, passing arguments to our target
                if self._target:
                    # bwcompat
                    kwargs['timedelta'] = timedelta

                    # TODO : use return code to determine when/how we need to run this the next time...
                    # Also we need to keep the exit status to be able to call external process as an update...

                    logging.debug("[{self.name}] calling {self._target.__name__} with args {args} and kwargs {kwargs}...".format(**locals()))
                    exitstatus = self._target(*args, **kwargs)

                if first_loop:
                    # signalling startup only at the end of the loop, only the first time
                    self.started.set()
                    first_loop = False

                if exitstatus is not None:
                    break

            if self.started.is_set() and exitstatus is None and self.exit.is_set():
                # in the not so special case where we started, we didnt get exit code and we exited,
                # this is expected as a normal result and we set an exitcode here of 0
                # As 0 is the conventional success for unix process successful run
                exitstatus = 0

        logging.debug("[{self.name}] Node stopped.".format(**locals()))
        return exitstatus  # returning last exit status from the update function

        # all context managers are destroyed properly here







# -*- coding: utf-8 -*-
from __future__ import absolute_import

# To allow python to run these tests as main script
import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..')))

import time
import multiprocessing
import pyzmp
import inspect

import pytest
# http://pytest.org/latest/contents.html
# https://github.com/ionelmc/pytest-benchmark

# TODO : PYPY
# http://pypy.org/


# IPC protocol
# Node as fixture to guarantee cleanup
# Better to have IPC as main class as it is simpler and easier to test than Socket.
class TestMockHWNodeIPC(object):
    __test__ = True

    class HWNode(pyzmp.Node):
        def __init__(self, name):
            super(TestMockHWNodeIPC.HWNode, self).__init__(name)
            self.magic_number = 666
            # TODO : improvement : autodetect class own methods
            # TODO : assert static ?
            self.provides(self.helloworld)
            self.provides(self.breakworld)
            self.provides(self.add)
            self.provides(self.getlucky)

        @staticmethod  # TODO : verify : is it true that a service is always a static method ( execution does not depend on instance <=> process local data ) ?
        def helloworld(msg):
            return "Hello! I am " + pyzmp.current_node().name if msg == "Hello" else "..."

        @staticmethod
        def breakworld(msg):
            raise Exception("Excepting Not Exceptionnally")

        @staticmethod
        def add(a, b):
            return a+b

        def getlucky(self):
            return self.magic_number

    def setup_method(self, method):
        # services is already setup globally
        self.hwnode = TestMockHWNodeIPC.HWNode(name="HNode")
        self.hwnodeextra = TestMockHWNodeIPC.HWNode(name="HNodeExtra")

    def teardown_method(self, method):
        if self.hwnode.is_alive():
            self.hwnode.shutdown(join=True)
        if self.hwnodeextra.is_alive():
            self.hwnodeextra.shutdown(join=True)
        # if it s still alive terminate it.
        if self.hwnode.is_alive():
            self.hwnode.terminate()
        if self.hwnodeextra.is_alive():
            self.hwnodeextra.terminate()

    # @nose.SkipTest  # to help debugging ( FIXME : how to programmatically start only one test - maybe in fixture - ? )
    def test_service_discover(self):
        print("\n" + inspect.currentframe().f_code.co_name)
        assert not self.hwnode.is_alive()

        print("Discovering helloworld Service...")
        helloworld = pyzmp.discover("helloworld")
        assert helloworld is None  # service not provided until node starts

        self.hwnode.start()
        assert self.hwnode.is_alive()

        print("Discovering helloworld Service...")
        helloworld = pyzmp.discover("helloworld", 5)  # we wait a bit to let it time to start
        assert not helloworld is None
        assert len(helloworld.providers) ==  1

        self.hwnode.shutdown()
        assert not self.hwnode.is_alive()

        print("Discovering helloworld Service...")
        helloworld = pyzmp.discover("helloworld")
        assert helloworld is None

    # @nose.SkipTest  # to help debugging ( FIXME : how to programmatically start only one test - maybe in fixture - ? )
    def test_service_discover_timeout(self):
        print("\n" + inspect.currentframe().f_code.co_name)
        assert not self.hwnode.is_alive()

        print("Discovering helloworld Service...")
        helloworld = pyzmp.discover("helloworld")
        assert helloworld is None  # service not provided until node starts

        print("Discovering helloworld Service...")
        helloworld = pyzmp.discover("helloworld", 1)  # check timeout actually times out
        assert helloworld is None

        print("Discovering helloworld Service...")
        helloworld = pyzmp.discover("helloworld", 1, 2)
        assert helloworld is None

    # @nose.SkipTest  # to help debugging ( FIXME : how to programmatically start only one test - maybe in fixture - ? )
    def test_service_discover_multiple_stack(self):
        print("\n" + inspect.currentframe().f_code.co_name)
        assert not self.hwnode.is_alive()

        print( "Discovering helloworld Service...")
        helloworld = pyzmp.discover("helloworld")
        assert helloworld is None  # service not provided until node starts

        # Start two nodes - stack process
        self.hwnodeextra.start()
        assert self.hwnodeextra.is_alive()

        print("Discovering helloworld Service...")
        helloworld = pyzmp.discover("helloworld", 5)  # we wait a bit to let it time to start
        assert not helloworld is None
        assert len(helloworld.providers) ==  1

        self.hwnode.start()
        assert self.hwnode.is_alive()

        print("Discovering helloworld Service...")
        helloworld = pyzmp.discover("helloworld", 5, 2)  # we wait until we get 2 providers ( or timeout )
        assert not helloworld is None
        assert len(helloworld.providers) ==  2

        self.hwnode.shutdown()
        assert not self.hwnode.is_alive()

        print("Discovering helloworld Service...")
        helloworld = pyzmp.discover("helloworld")  # we should have right away 1 provider only
        assert not helloworld is None
        assert len(helloworld.providers) ==  1

        self.hwnodeextra.shutdown()
        assert not self.hwnodeextra.is_alive()

    # @nose.SkipTest  # to help debugging ( FIXME : how to programmatically start only one test - maybe in fixture - ? )
    def test_service_discover_multiple_queue(self):
        print("\n" + inspect.currentframe().f_code.co_name)
        assert not self.hwnode.is_alive()

        print("Discovering helloworld Service...")
        helloworld = pyzmp.discover("helloworld")
        assert helloworld is None  # service not provided until node starts

        # Start two nodes queue process
        self.hwnodeextra.start()
        assert self.hwnodeextra.is_alive()

        print("Discovering helloworld Service...")
        helloworld = pyzmp.discover("helloworld", 5)  # we wait a bit to let it time to start
        assert not helloworld is None
        assert len(helloworld.providers) ==  1

        self.hwnode.start()
        assert self.hwnode.is_alive()

        print("Discovering helloworld Service...")
        helloworld = pyzmp.discover("helloworld", 5, 2)   # we wait until we get 2 providers ( or timeout )
        assert not helloworld is None
        assert len(helloworld.providers) ==  2

        self.hwnodeextra.shutdown()
        assert not self.hwnodeextra.is_alive()

        print("Discovering helloworld Service...")
        helloworld = pyzmp.discover("helloworld", 5)  # we wait a bit to let it time to start
        assert not helloworld is None
        assert len(helloworld.providers) ==  1

        self.hwnode.shutdown()
        assert not self.hwnode.is_alive()

    # @nose.SkipTest  # to help debugging ( FIXME : how to programmatically start only one test - maybe in fixture - ? )
    def test_service_comm_to_sub(self):

        print("\n" + inspect.currentframe().f_code.co_name)
        assert not self.hwnode.is_alive()
        self.hwnode.start()
        assert self.hwnode.is_alive()

        print("Discovering helloworld Service...")
        helloworld = pyzmp.discover("helloworld", 5)
        assert helloworld is not None  # to make sure we get a service provided
        resp = helloworld.call(args=("Hello",))
        print("Hello -> {0}".format(resp))
        assert resp == "Hello! I am HNode"
        resp = helloworld.call(args=("Hallo",))
        print("Hallo -> {0}".format(resp))
        assert resp == "..."

        self.hwnode.shutdown()
        assert not self.hwnode.is_alive()

    # @nose.SkipTest  # to help debugging ( FIXME : how to programmatically start only one test - maybe in fixture - ? )
    def test_service_comm_to_double_sub(self):
        print("\n" + inspect.currentframe().f_code.co_name)

        assert not self.hwnode.is_alive()
        self.hwnode.start()
        assert self.hwnode.is_alive()

        assert not self.hwnodeextra.is_alive()
        self.hwnodeextra.start()
        assert self.hwnodeextra.is_alive()

        print("Discovering helloworld Service...")
        helloworld = pyzmp.discover("helloworld", 5, 2)  # make sure we get both providers. we need them.
        assert helloworld is not None  # to make sure we get a service provided
        assert len(helloworld.providers) ==  2
        resp = helloworld.call(args=("Hello",))
        print("Hello -> {0}".format(resp))
        assert resp == "Hello! I am HNode" or resp == "Hello! I am HNodeExtra"
        resp = helloworld.call(args=("Hallo",))
        print("Hallo -> {0}".format(resp))
        assert resp == "..."

        resp = helloworld.call(args=("Hello",), node=self.hwnode.name)
        print("Hello -HNode-> {0}".format(resp))
        assert resp == "Hello! I am HNode"
        resp = helloworld.call(args=("Hello",), node=self.hwnodeextra.name)
        print("Hello -HNodeExtra-> {0}".format(resp))
        assert resp == "Hello! I am HNodeExtra"

        self.hwnode.shutdown()
        assert not self.hwnode.is_alive()
        self.hwnodeextra.shutdown()
        assert not self.hwnodeextra.is_alive()

    # @nose.SkipTest  # to help debugging ( FIXME : how to programmatically start only one test - maybe in fixture - ? )
    def test_service_double_comm_to_sub(self):

        print("\n" + inspect.currentframe().f_code.co_name)
        assert not self.hwnode.is_alive()
        self.hwnode.start()
        assert self.hwnode.is_alive()

        print("Discovering helloworld Service...")
        helloworld = pyzmp.discover("helloworld", 5)
        assert helloworld is not None  # to make sure we get a service provided

        def callit():
            hw = pyzmp.discover("helloworld", 5)
            return hw.call(args=("Hello",))

        c = multiprocessing.Process(name="Client", target=callit)
        assert not c.is_alive()
        c.start()
        assert c.is_alive()

        resp = helloworld.call(args=("Hello",))
        print("Hallo -> {0}".format(resp))
        assert resp == "Hello! I am HNode"
        resp = helloworld.call(args=("Hallo",))
        print("Hallo -> {0}".format(resp))
        assert resp == "..."
        resp = helloworld.call(args=("Hello",), node=self.hwnode.name)
        print("Hello -HNode-> {0}".format(resp))
        assert resp == "Hello! I am HNode"

        c.join()
        assert not c.is_alive()

        self.hwnode.shutdown()
        assert not self.hwnode.is_alive()

    # @nose.SkipTest  # to help debugging ( FIXME : how to programmatically start only one test - maybe in fixture - ? )
    def test_service_comm_to_sub_self(self):

        print("\n" + inspect.currentframe().f_code.co_name)
        assert not self.hwnode.is_alive()
        self.hwnode.magic_number = 42
        self.hwnode.start()
        assert self.hwnode.is_alive()

        print("Discovering getlucky Service...")
        getlucky = pyzmp.discover("getlucky", 5)
        assert getlucky is not None  # to make sure we get a service provided
        resp = getlucky.call()
        print("42 ? -> {0}".format(resp))
        assert resp == 42

        self.hwnode.shutdown()
        assert not self.hwnode.is_alive()

    # @nose.SkipTest  # to help debugging ( FIXME : how to programmatically start only one test - maybe in fixture - ? )
    def test_service_comm_to_double_sub_self(self):
        print("\n" + inspect.currentframe().f_code.co_name)

        assert not self.hwnode.is_alive()
        self.hwnode.magic_number = 42
        self.hwnode.start()
        assert self.hwnode.is_alive()

        assert not self.hwnodeextra.is_alive()
        self.hwnodeextra.magic_number = 79
        self.hwnodeextra.start()
        assert self.hwnodeextra.is_alive()

        print("Discovering getlucky Service...")
        getlucky = pyzmp.discover("getlucky", 5, 2)  # make sure we get both providers. we need them.
        assert getlucky is not None  # to make sure we get a service provided
        assert len(getlucky.providers) ==  2
        resp = getlucky.call()
        print("42 || 79 ? -> {0}".format(resp))
        assert resp == 42 or resp == 79

        resp = getlucky.call(node=self.hwnode.name)
        print("42 ? -HNode-> {0}".format(resp))
        assert resp == 42
        resp = getlucky.call(node=self.hwnodeextra.name)
        print("79 ? -HNodeExtra-> {0}".format(resp))
        assert resp == 79

        self.hwnode.shutdown()
        assert not self.hwnode.is_alive()
        self.hwnodeextra.shutdown()
        assert not self.hwnodeextra.is_alive()

    # @nose.SkipTest  # to help debugging ( FIXME : how to programmatically start only one test - maybe in fixture - ? )
    def test_service_double_comm_to_sub_self(self):

        print("\n" + inspect.currentframe().f_code.co_name)
        assert not self.hwnode.is_alive()
        self.hwnode.magic_number = 42
        self.hwnode.start()
        assert self.hwnode.is_alive()

        print("Discovering getlucky Service...")
        getlucky = pyzmp.discover("getlucky", 5)
        assert getlucky is not None  # to make sure we get a service provided

        def callit():
            getlucky = pyzmp.discover("getlucky", 5)
            return getlucky.call()

        c = multiprocessing.Process(name="Client", target=callit)
        assert not c.is_alive()
        c.start()
        assert c.is_alive()

        resp = getlucky.call()
        print("42 ? -> {0}".format(resp))
        assert resp == 42
        resp = getlucky.call(node=self.hwnode.name)
        print("42 ? -HNode-> {0}".format(resp))
        assert resp == 42

        c.join()
        assert not c.is_alive()

        self.hwnode.shutdown()
        assert not self.hwnode.is_alive()



    # @nose.SkipTest  # to help debugging ( FIXME : how to programmatically start only one test - maybe in fixture - ? )
    def test_service_except_from_sub(self):

        print("\n" + inspect.currentframe().f_code.co_name)
        assert not self.hwnode.is_alive()
        self.hwnode.start()
        assert self.hwnode.is_alive()

        print("Discovering breakworld Service...")
        breakworld = pyzmp.discover("breakworld", 5)
        assert breakworld is not None  # to make sure we get a service provided
        with pytest.raises(Exception) as cm:
            resp = breakworld.call(args=("Hello",))

        self.hwnode.shutdown()
        assert not self.hwnode.is_alive()

    # @nose.SkipTest  # to help debugging ( FIXME : how to programmatically start only one test - maybe in fixture - ? )
    def test_service_except_from_node_no_service(self):

        print("\n" + inspect.currentframe().f_code.co_name)
        assert not self.hwnode.is_alive()
        self.hwnode.start()
        assert self.hwnode.is_alive()

        print("Discovering breakworld Service...")
        breakworld = pyzmp.discover("breakworld", 5)
        assert breakworld is not None  # to make sure we get a service provided

        # messing around even if we should not
        breakworld.name = "NOT_EXISTING"

        with pytest.raises(pyzmp.UnknownServiceException) as cm:
            resp = breakworld.call(args=("Hello",))

        self.hwnode.shutdown()
        assert not self.hwnode.is_alive()

    # TODO : check mo exception cases

    # @nose.SkipTest  # to help debugging ( FIXME : how to programmatically start only one test - maybe in fixture - ? )
    def test_service_params_comm_to_sub(self):

        print("\n" + inspect.currentframe().f_code.co_name)
        assert not self.hwnode.is_alive()
        self.hwnode.start()
        assert self.hwnode.is_alive()

        print("Discovering add Service...")
        add = pyzmp.discover("add", 5)
        assert add is not None  # to make sure we get a service provided

        resp = add.call(args=(17, 25))
        print(" 17 + 25 -> {0}".format(resp))
        assert resp == 17+25

        self.hwnode.shutdown()
        assert not self.hwnode.is_alive()


# Node as fixture to guarantee cleanup
# TCP protocol
class TestMockHWNodeSocket(TestMockHWNodeIPC):
    __test__ = True

    class HWNode(pyzmp.Node):
        def __init__(self, name, socket_bind):
            super(TestMockHWNodeSocket.HWNode, self).__init__(name, socket_bind)
            self.magic_number = 999
            # TODO : improvement : autodetect class own methods
            # TODO : assert static ?
            self.provides(self.helloworld)
            self.provides(self.breakworld)
            self.provides(self.add)
            self.provides(self.getlucky)

        @staticmethod  # TODO : verify : is it true that a service is always a static method ( execution does not depend on instance <=> process local data ) ?
        def helloworld(msg):
            return "Hello! I am " + pyzmp.current_node().name if msg == "Hello" else "..."

        @staticmethod
        def breakworld(msg):
            raise Exception("Excepting Not Exceptionnally")

        @staticmethod
        def add(a, b):
            return a+b

        def getlucky(self):
            return self.magic_number

    def setup_method(self, method):
        # services is already setup globally
        self.hwnode = TestMockHWNodeSocket.HWNode(name="HNode", socket_bind="tcp://127.0.0.1:4242")
        self.hwnodeextra = TestMockHWNodeSocket.HWNode(name="HNodeExtra", socket_bind="tcp://127.0.0.1:4243")

    def teardown_method(self, method):
        if self.hwnode.is_alive():
            self.hwnode.shutdown(join=True)
        if self.hwnodeextra.is_alive():
            self.hwnodeextra.shutdown(join=True)
        # if it s still alive terminate it.
        if self.hwnode.is_alive():
            self.hwnode.terminate()
        if self.hwnodeextra.is_alive():
            self.hwnodeextra.terminate()

    #TODO : verify all tests are run again with proper fixture





# -*- coding: utf-8 -*-
# This python package is implementing a simple multiprocess framework.
# Although we need to keep it simple, and minimalist, it should provide everything needed
# for efficient multiprocess programming.
# The goal is to make is as simple as possible to build resilient parallel software, like with functional programming,
# yet to be in total control of what each process does.
# Possible target applications are multi process systems ( ROS, HPC, celery, Neural Networks, etc. )
#

# Concepts to keep in mind :
# - services (request based - pull oriented model - sync or async - Ex: web, RPC, etc.)
# - topic/feeds ( message based - push oriented model - sync or async - Ex: Pub/Sub, XMPP, etc.)
# - echo nodes ( only echo back messages or requests as is)
# - forward nodes ( only forward messages or requests as is)
# - shadows ( allow switching from push model to pull model via an intermediate that store data - cf. Amazon IoT architecture design )
# - ghost ( allow switching from pull model to push model via an intermediate that broadcast estimated data )
# ghost service depends on shadow service, shadow feed depends on ghost feed... kind of.
# Lots of explicit details are needed in implementation ( frequency of publishing, paramters of requests, variation of data/response, etc.)
# Estimating/simulating all that is not trivial, but that's very likely the gist of the "learning" behavior of neuronal networks.

# In pyros, the point of it is to be able to fully tests the multiprocess behavior,
#     in pure python, without having to run a ROS system.
from __future__ import absolute_import

from .exceptions import UnknownServiceException, UnknownRequestTypeException, UnknownResponseTypeException
from .master import manager
from .node import Node, current_node
from .service import Service, services, discover, ServiceCallTimeout


topics = manager.list()
params = manager.list()


from .helpers import process_watcher
from .helpers import process_watcher_cm


__all__ = [
    'UnknownServiceException', 'UnknownRequestTypeException', 'UnknownResponseTypeException',
    'Node', 'nodes', 'current_node',
    'Service', 'services', 'discover', 'ServiceCallTimeout'
]



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

import logging

# Protobuf is used if message has been generated.
# Otherwise use pickleable fallback
# Note : Many protobuf methods do not have implementation in pickle protocol.
_PROTOBUF = False

# Only for fallback behavior
import pickle
from collections import namedtuple

# TODO : this should probably in pyros, not in zmp...

try:
    from .service_pb2 import ServiceRequestImpl, ServiceResponseImpl, ServiceExceptionImpl
    _PROTOBUF = True
except (ImportError):
    logging.warn("ZMQ : Protobuf message implementation not found. Using pickle based protocol")
    # Note : pickleable interface is inferior since type is not checked on assignment.
    # however the functional pickle interface is better than the object oriented protobuf message interface.
    # Another difference is that namedtuple are readonly, which is okay in most cases.
    #
    # The pickleable interface might just be a useful fallback if you don't want to use protobuf.
    # In that case, just call force_namedtuple_implementation() after importing this module to override default setup.

protobuf_implementation_enabled = False


#TODO : check python3 to avoid too much hacking here
def force_protobuf_implementation():

    if not _PROTOBUF:
        logging.error("Cannot activate protobuf implementation. Protobuf package not imported.")
        return False

    global protobuf_implementation_enabled
    protobuf_implementation_enabled = True
    # Dynamic Functional Facade for internal message format

    global ServiceRequest, ServiceRequest_dictparse
    # Note : isinstance returns True for both class names
    ServiceRequest = ServiceRequestImpl
    ServiceRequest.initialized = ServiceRequestImpl.IsInitialized
    ServiceRequest.has_field = lambda s, f: s.HasField(f)
    ServiceRequest.serialize = ServiceRequestImpl.SerializeToString
    ServiceRequest._parse = lambda s, m: [s, s.ParseFromString(m)][0]  # returns itself after modification during parsing
    # Factory Function for oneline parsing creation
    ServiceRequest_dictparse = lambda m: ServiceRequest(**{f.name: v for (f, v) in ServiceRequest()._parse(m).ListFields()})

    global ServiceResponse, ServiceResponse_dictparse
    # Note : isinstance returns True for both class names
    ServiceResponse = ServiceResponseImpl
    ServiceResponse.initialized = ServiceResponseImpl.IsInitialized
    ServiceResponse.has_field = lambda s, f: s.HasField(f)
    ServiceResponse.serialize = ServiceResponseImpl.SerializeToString
    ServiceResponse._parse = lambda s, m: [s, s.ParseFromString(m)][0]  # returns itself after modification during parsing
    # Factory Function for oneline parsing creation
    ServiceResponse_dictparse = lambda m: ServiceResponse(**{f.name: v for (f, v) in ServiceResponse()._parse(m).ListFields()})
    #TODO : check if we can use CopyFrom() instead ?

    global ServiceException, ServiceException_dictparse
    # Note : isinstance returns True for both class names
    ServiceException = ServiceExceptionImpl
    ServiceException.initialized = ServiceExceptionImpl.IsInitialized
    ServiceException.has_field = lambda s, f: s.HasField(f)
    ServiceException.serialize = ServiceExceptionImpl.SerializeToString
    ServiceException._parse = lambda s, m: [s, s.ParseFromString(m)][0]  # returns itself after modification during parsing
    # Factory Function for oneline parsing creation
    ServiceException_dictparse = lambda m: ServiceException(**{f.name: v for (f, v) in ServiceException()._parse(m).ListFields()})

    # TODO : Generic / Template class / Functional Interface for any kind of message


# Fallback implementation. need to be defined at module level to be pickleable ( unless dill is used )
ServiceRequestNTImpl = namedtuple("ServiceRequest", "service args kwargs")
ServiceResponseNTImpl = namedtuple("ServiceResponse", "service response exception")
ServiceExceptionNTImpl = namedtuple("ServiceException", "exc_type exc_value traceback")


def force_namedtuple_implementation():
    global protobuf_implementation_enabled
    protobuf_implementation_enabled = False
    # Dynamic Functional Facade for internal message format

    global ServiceRequest, ServiceRequest_dictparse
    # Extend named tuple implementation
    ServiceRequest = ServiceRequestNTImpl
    ServiceRequest.__new__.__defaults__ = (None,) * len(ServiceRequest._fields)  # making all fields optional ( like in protobuf protocol )
    ServiceRequest.initialized = lambda s: True
    ServiceRequest.has_field = lambda s, f: getattr(s, f, None) is not None
    ServiceRequest.serialize = lambda s: pickle.dumps(s)
    # Factory Function for oneline parsing creation
    ServiceRequest_dictparse = lambda m: pickle.loads(m)

    global ServiceResponse, ServiceResponse_dictparse
    # Extend named tuple implementation
    ServiceResponse = ServiceResponseNTImpl
    ServiceResponse.__new__.__defaults__ = (None,) * len(ServiceResponse._fields)  # making all fields optional ( like in protobuf protocol )
    ServiceResponse.initialized = lambda s: True
    ServiceResponse.has_field = lambda s, f: getattr(s, f, None) is not None
    ServiceResponse.serialize = lambda s: pickle.dumps(s)
    # Factory Function for oneline parsing creation
    ServiceResponse_dictparse = lambda m: pickle.loads(m)

    global ServiceException, ServiceException_dictparse
    # Extend named tuple implementation
    ServiceException = ServiceExceptionNTImpl
    ServiceException.__new__.__defaults__ = (None,) * len(ServiceException._fields)  # making all fields optional ( like in protobuf protocol )
    ServiceException.initialized = lambda s: True
    ServiceException.has_field = lambda s, f: getattr(s, f, None) is not None
    ServiceException.serialize = lambda s: pickle.dumps(s)
    # Factory Function for oneline parsing creation
    ServiceException_dictparse = lambda m: pickle.loads(m)

    # TODO : Generic / Template class / Functional Interface for any kind of message


# Default behavior : choose protobuf if available
if _PROTOBUF:
    force_protobuf_implementation()
else:
    force_namedtuple_implementation()





from __future__ import absolute_import
from __future__ import print_function

import sys


def main():
    if len(sys.argv) > 1 and sys.argv[1] == '--pytest':
        import pytest  # import only if needed
        errno = pytest.main(sys.argv[2:])
        sys.exit(errno)


if __name__ == "__main__":
    main()




from __future__ import absolute_import
from __future__ import print_function

import contextlib
import os
import multiprocessing
import subprocess

# TODO : turn this into an actual pyzmp Node.
# It will also give us a way to encapsulate shell commands in a pyzmp node, which will be useful.

# TODO : mix all this with xonsh ?

# TODO :check psutil for alternative implementation


def watcher_proc(pid=None):
    """
    Simply encapsulate a shell call to pstree into python, the standard way.
    :param pid:
    :return:
    """
    # : this is a synchronous call that waits for completion
    return subprocess.check_call("xterm -e watch pstree -cap {0}".format(str(pid)), shell=True)


def process_watcher(pid=None, daemon=True):
    pid = pid or os.getpid()
    watchp = multiprocessing.Process(target=watcher_proc, name="node_watcher", args=(pid,))
    watchp.daemon = daemon  # make the process daemonic to guarantee it is terminated when parent terminates
    watchp.start()
    return watchp


@contextlib.contextmanager
def process_watcher_cm(pid=None, daemon=True):
    pid = pid or os.getpid()
    watchp = multiprocessing.Process(target=watcher_proc, name="node_watcher", args=(pid,))
    watchp.daemon = daemon  # make the process daemonic to guarantee it is terminated when parent terminates
    watchp.start()
    try:
        yield watchp
    finally:
        watchp.terminate()



import os
import shutil
import subprocess
import sys
import tempfile
import importlib

import setuptools

# Ref : https://packaging.python.org/single_source_version/#single-sourcing-the-version
with open('pyzmp/_version.py') as vf:
    exec(vf.read())

# Best Flow :
# Clean previous build & dist
# $ gitchangelog >CHANGELOG.rst
# change version in code and changelog
# $ python setup.py prepare_release
# WAIT FOR TRAVIS CHECKS
# $ python setup.py publish
# => TODO : try to do a simpler "release" command


# Clean way to add a custom "python setup.py <command>"
# Ref setup.py command extension : https://blog.niteoweb.com/setuptools-run-custom-code-in-setup-py/
class PrepareReleaseCommand(setuptools.Command):
    """Command to release this package to Pypi"""
    description = "prepare a release of pyros"
    user_options = []

    def initialize_options(self):
        """init options"""
        pass

    def finalize_options(self):
        """finalize options"""
        pass

    def run(self):
        """runner"""

        # TODO :
        # $ gitchangelog >CHANGELOG.rst
        # $ git commit CHANGELOG.rst -m "updating changelog"
        # change version in code and changelog
        subprocess.check_call("git commit CHANGELOG.rst pyzmp/_version.py -m 'v{0}'".format(__version__), shell=True)
        subprocess.check_call("git push", shell=True)

        print("You should verify travis checks, and you can publish this release with :")
        print("  python setup.py publish")
        sys.exit()


# Clean way to add a custom "python setup.py <command>"
# Ref setup.py command extension : https://blog.niteoweb.com/setuptools-run-custom-code-in-setup-py/
class PublishCommand(setuptools.Command):
    """Command to release this package to Pypi"""
    description = "releases pyros to Pypi"
    user_options = []

    def initialize_options(self):
        """init options"""
        pass

    def finalize_options(self):
        """finalize options"""
        pass

    def run(self):
        """runner"""

        subprocess.check_call("python setup.py sdist", shell=True)
        subprocess.check_call("python setup.py bdist_wheel", shell=True)
        # OLD way:
        # os.system("python setup.py sdist bdist_wheel upload")
        # NEW way:
        # Ref: https://packaging.python.org/distributing/
        subprocess.check_call("twine upload dist/*", shell=True)

        subprocess.check_call("git tag -a {0} -m 'version {0}'".format(__version__), shell=True)
        subprocess.check_call("git push --tags", shell=True)
        sys.exit()


# Clean way to add a custom "python setup.py <command>"
# Ref setup.py command extension : https://blog.niteoweb.com/setuptools-run-custom-code-in-setup-py/
class RosDevelopCommand(setuptools.Command):

    """Command to mutate this package to a ROS package, using its ROS release repository"""
    description = "mutate this package to a ROS package, using its ROS release repository"
    user_options = []

    def initialize_options(self):
        """init options"""
        # TODO : add distro selector
        pass

    def finalize_options(self):
        """finalize options"""
        pass

    def run(self):
        # dynamic import for this command only to not need these in usual python case...
        import git
        import yaml

        """runner"""
        repo_path = tempfile.mkdtemp(prefix='rosdevelop-' + os.path.dirname(__file__))  # TODO get actual package name ?
        print("Getting ROS release repo in {0}...".format(repo_path))
        rosrelease_repo = git.Repo.clone_from('https://github.com/asmodehn/pyzmp-rosrelease.git', repo_path)

        # Reset our working tree to master
        origin = rosrelease_repo.remotes.origin
        rosrelease_repo.remotes.origin.fetch()  # assure we actually have data. fetch() returns useful information
        # Setup a local tracking branch of a remote branch
        rosrelease_repo.create_head('master', origin.refs.master).set_tracking_branch(origin.refs.master).checkout()

        print("Reading tracks.yaml...")
        with open(os.path.join(rosrelease_repo.working_tree_dir, 'tracks.yaml'), 'r') as tracks:
            try:
                tracks_dict = yaml.load(tracks)
            except yaml.YAMLError as exc:
                raise

        patch_dir = tracks_dict.get('tracks', {}).get('indigo', {}).get('patches', {})

        print("Found patches for indigo in {0}".format(patch_dir))
        src_files = os.listdir(os.path.join(rosrelease_repo.working_tree_dir, patch_dir))

        working_repo = git.Repo(os.path.dirname(os.path.abspath(__file__)))

        # adding patched files to ignore list if needed (to prevent accidental commit of patch)
        # => BETTER if the patch do not erase previous file. TODO : fix problem with both .travis.yml
        with open(os.path.join(working_repo.working_tree_dir, '.gitignore'), 'a+') as gitignore:
            skipit = []
            for line in gitignore:
                if line in src_files:
                    skipit += line
                else:  # not found, we are at the eof
                    for f in src_files:
                        if f not in skipit:
                            gitignore.write(f+'\n')  # append missing data

        working_repo.git.add(['.gitignore'])  # adding .gitignore to the index so git applies it (and hide new files)

        for file_name in src_files:
            print("Patching {0}".format(file_name))
            full_file_name = os.path.join(rosrelease_repo.working_tree_dir, patch_dir, file_name)
            if os.path.isfile(full_file_name):
                # Special case for package.xml and version template string
                if file_name == 'package.xml':
                    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'package.xml'), "wt") as fout:
                        with open(full_file_name, "rt") as fin:
                            for line in fin:
                                fout.write(line.replace(':{version}', __version__))  # TODO: proper template engine ?
                else:
                    shutil.copy(full_file_name, os.path.dirname(os.path.abspath(__file__)))

        sys.exit()


setuptools.setup(name='pyzmp',
    version=__version__,
    description='ZeroMq based multiprocessing framework.',
    url='http://github.com/asmodehn/pyzmp',
    author='AlexV',
    author_email='asmodehn@gmail.com',
    license='BSD',
    packages=[
        'pyzmp',
        'pyzmp.tests'
    ],
    entry_points={
        'console_scripts': [
            'pyzmp = pyzmp.__main__:main'
        ]
    },
    # this is better than using package data ( since behavior is a bit different from distutils... )
    include_package_data=True,  # use MANIFEST.in during install.
    install_requires=[
        'tblib',  # this might not always install six (latest version does not)
        'six',
        'pyzmq',
        'pytest-timeout',
        # Careful : upon install plugins can be resolved instead of core pytest package
        # => pytest should be listed last here...
        'pytest>=2.9.1',  # since tests are embedded in package
    ],
    setup_requires=['pytest-runner'],
    tests_require=['pytest>=2.9.1'],
    cmdclass={
        'rosdevelop': RosDevelopCommand,
        'prepare_release': PrepareReleaseCommand,
        'publish': PublishCommand,
    },
    zip_safe=False,  # TODO testing...
)
